/***************************************************************************
 *            graphics.c
 *
 *  Tue Dec 21 15:40:34 2004
 *  Copyright  2004  Jean-Pierre Lozi
 *  jean-pierre@lozi.org - http://gtuxball.lozi.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include <stdio.h>
#include <time.h>

/* We need to include the gtk headers... */
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>

/* ...and our local headers. */
#include "graphics.h"
#include "element.h"
#include "game.h"
#include "globals.h"
#include "score.h"
#include "error.h"

#define EMPTY_CURSOR_WIDTH 16
#define EMPTY_CURSOR_HEIGHT 16

/* Those two local variables represent an empty cursor, which will be
 * used when the mouse enters the graphic area. We could of course use
 * a for loop, or create a 1-pixel-wide cursor, but declaring it that
 * way allows us to easily change the cursor. */
const static unsigned char empty_cursor_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
  
const static unsigned char empty_cursor_mask_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

/* Private functions : */

/* This function initializes the menus. */
void graphics_init_menus ( GtkBox * parent, GtkWindow * window );

/* This functions allows to quickly create a menu item. */
GtkWidget *graphics_create_menu_item( gchar * caption, gchar * accel,
									  gchar * icon, GtkWidget * menu,
									  GCallback callback, gpointer data,
									  gboolean radio_menu_item,
									  GSList * list );

/* This functions allows to quickly create a menu separator. */
GtkWidget *graphics_create_menu_separator( GtkWidget * menu );

/* This function is called when the window is destroyed. */
void graphics_on_destroy ( GtkWidget * widget, gpointer data );

/* This function is called when the window has to be redrawn. */
gboolean graphics_on_expose_event ( GtkWidget *widget, GdkEventExpose *event,
									gpointer data);

/* This function is called when the user presses a key. */
gboolean graphics_on_key_press_event( GtkWidget *widget, GdkEventKey *event,
                                      gpointer user_data);
									  
/* This function is called when the mouse moves. */
gboolean graphics_on_motion_notify_event (	GtkWidget *widget, GdkEventMotion *event,
								 			gpointer data );

/* This function is called when the mouse enters the drawing area. */
gboolean graphics_on_enter_notify_event ( GtkWidget *widget,
										  GdkEventCrossing *event,
										  gpointer data );

/* This function is called when the mouse leaves the drawing area. */
gboolean graphics_on_leave_notify_event ( GtkWidget *widget,
										  GdkEventCrossing *event,
										  gpointer data );

/* This function is called when the user chooses "New Game" from the Game menu. */
void graphics_on_new_game ( GtkMenuItem *widget, gpointer user_data );

/* This function is called when the user chooses "Pause Game" from the Game menu. */
void graphics_on_pause_game ( GtkMenuItem *widget, gpointer user_data );

/* This function is called when any of the "Scores" menu item is activated. */
void graphics_on_scores( GtkMenuItem *menuitem, gpointer data );

/* This function is called when the user chooses "End Game" from the Game menu. */
void graphics_on_end_game ( GtkMenuItem *widget, gpointer user_data);

/* This function is called when any of the "Level" menu items is activated. */
void graphics_on_level( GtkMenuItem *menuitem, gpointer data );

/* This function is called when the users chooses "About" in the help menu. */
void graphics_on_about( GtkMenuItem *widget, gpointer user_data );

/* This function wraps the real timeout function (in order to avoid to use GTK
 * types in game.c). */
gboolean graphics_timeout_function( gpointer data );

/* This function draws the given message. */
void graphics_draw_message( char * message, GdkGC * gc, GtkWidget * widget,
							int display_infos );


/*
 * Global variables
 */
GtkWidget *drawing_area, *window, *pause_game, *end_game;
GdkPixbuf	*pad1_left_pixbuf, *pad2_left_pixbuf, *pad1_right_pixbuf, *pad2_right_pixbuf,
			*ball_pixbuf, *background_pixbuf, *blocks_pixbufs[8];
GdkCursor *arrow_cursor, *empty_cursor;
gpointer levels;

/*
 * This function initializes the window and starts the event loop.
 */
void graphics_init (int argc, char *argv[])
{
	GtkWidget *vbox;
	GdkColor black = { 0, 0, 0, 0 };
	int i;
	
	/* We first call gtk_init. */
    gtk_init (&argc, &argv);

	/*
	 * We load all the pictures into memory:
	 */
	/* The background picture... */
	background_pixbuf = gtk_image_get_pixbuf( 
						GTK_IMAGE(gtk_image_new_from_file( BACKGROUND_IMAGE_FILE ) ) );
	
	/* ...the pad... */
	pad1_left_pixbuf = gtk_image_get_pixbuf( 
			GTK_IMAGE(gtk_image_new_from_file( pad_image_files[PAD1_LEFT] ) ) );
	pad2_left_pixbuf = gtk_image_get_pixbuf(
			GTK_IMAGE(gtk_image_new_from_file( pad_image_files[PAD2_LEFT] ) ) );
	pad1_right_pixbuf = gtk_image_get_pixbuf(
			GTK_IMAGE(gtk_image_new_from_file( pad_image_files[PAD1_RIGHT] ) ) );
	pad2_right_pixbuf = gtk_image_get_pixbuf(
			GTK_IMAGE(gtk_image_new_from_file( pad_image_files[PAD2_RIGHT] ) ) );
	
	/* ...the ball... */
	ball_pixbuf = gtk_image_get_pixbuf( GTK_IMAGE(gtk_image_new_from_file( ball_image_file ) ) );
	
	/* ...and each block. */
	for( i = 1 ; i <= NUMBER_OF_COLORS ; i++ )
	{
		blocks_pixbufs[i] = gtk_image_get_pixbuf(
							  GTK_IMAGE(gtk_image_new_from_file( blocks_image_files[i - 1] ) ) );
	}
	
	/*
	 * We load all the cursors into memory:
	 */
	/* The empty cursor... */
	empty_cursor = gdk_cursor_new_from_pixmap( 
					gdk_bitmap_create_from_data ( NULL, empty_cursor_bits,
                                  				  EMPTY_CURSOR_WIDTH,
												  EMPTY_CURSOR_HEIGHT ),
					gdk_bitmap_create_from_data (NULL, empty_cursor_mask_bits,
                                  				  EMPTY_CURSOR_WIDTH,
												  EMPTY_CURSOR_HEIGHT ),
					&black, &black, 8, 8);
	
	/* ...and the arrow cursor. */
	arrow_cursor = gdk_cursor_new( GDK_ARROW );
	
	/* Then, we create a new window... */
    window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_policy(GTK_WINDOW(window), 0, 0, 0);
    gtk_container_set_border_width (GTK_CONTAINER (window), 5);
    gtk_window_set_title (GTK_WINDOW (window), "gTuxBall");
	
	/* ...connect the signals... */
    g_signal_connect (G_OBJECT (window), "destroy",
                      G_CALLBACK (graphics_on_destroy), NULL);
	
	/* ...create a vbox...*/
    vbox = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(window), vbox);
	
	/* ...create menus... */
	graphics_init_menus(GTK_BOX( vbox ), GTK_WINDOW(window));

    /* ...create the drawing area... */
	drawing_area = gtk_drawing_area_new();
	gtk_widget_set_size_request(GTK_WIDGET(drawing_area),645,480);
	gtk_box_pack_end(GTK_BOX(vbox), drawing_area, FALSE, FALSE, 2);
	
	/* ...connect the right signals... */
	gtk_widget_add_events ( GTK_WIDGET(drawing_area), 
							GDK_POINTER_MOTION_MASK | GDK_ENTER_NOTIFY_MASK
							| GDK_LEAVE_NOTIFY_MASK | GDK_POINTER_MOTION_HINT_MASK );
	
	g_signal_connect ( G_OBJECT (drawing_area), "expose_event",  
                       G_CALLBACK (graphics_on_expose_event), NULL );
	g_signal_connect ( G_OBJECT (drawing_area), "motion_notify_event",  
                       G_CALLBACK (graphics_on_motion_notify_event), NULL );
	g_signal_connect ( G_OBJECT (drawing_area), "enter_notify_event",  
                       G_CALLBACK (graphics_on_enter_notify_event), NULL );
	g_signal_connect ( G_OBJECT (drawing_area), "leave_notify_event",  
                       G_CALLBACK (graphics_on_leave_notify_event), NULL );
	
	/* ...show all the widgets... */		
	gtk_widget_show_all (window);
	
	g_signal_connect (G_OBJECT (window), "key_press_event",  
                      G_CALLBACK (graphics_on_key_press_event), drawing_area);
	
	/* ...and start the event loop. */
    gtk_main ();
	
	return;
}

/*
 * This function initializes the menus.
 */
void graphics_init_menus ( GtkBox * parent, GtkWindow * window )
{
	/* We first declare all the variables we need... */
	GtkWidget 	*menu_bar, *game_menu, *settings_menu, *help_menu, *menu,
				*menu_item, *level_submenu;
	GtkAccelGroup *accel_group = gtk_accel_group_new();
	GSList * group = NULL;
	int i;
	
	/* ...then create the menu bar... */
	menu_bar = gtk_menu_bar_new();
	gtk_box_pack_start(GTK_BOX(parent), menu_bar, FALSE, FALSE, 2);
	
	/*
	 * ...create each menu:
	 */
	/*=======================================================================*/
	/* Game menu                                                             */
	/*=======================================================================*/
	game_menu = gtk_menu_item_new_with_mnemonic( "_Game" );
	
	
	menu = gtk_menu_new();
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(game_menu), GTK_WIDGET(menu));
	
	/* AccelGroup / AccelMap creation. */
	gtk_window_add_accel_group (GTK_WINDOW (window), accel_group);
	gtk_menu_set_accel_group (GTK_MENU (menu), accel_group);
	gtk_accel_map_add_entry (	"<gTuxBall-Main>/Game/New Game",
								gdk_keyval_from_name ("n"),
								GDK_CONTROL_MASK);
	gtk_accel_map_add_entry (	"<gTuxBall>/Game/Pause Game",
								GDK_Pause,
								0);
	gtk_accel_map_add_entry (	"<gTuxBall>/Game/Quit",
								gdk_keyval_from_name ("q"),
								GDK_CONTROL_MASK);
	
	menu_item = graphics_create_menu_item (	"_New Game",
											"<gTuxBall-Main>/Game/New Game",
											NULL, menu,
											G_CALLBACK(&graphics_on_new_game), NULL, FALSE, NULL );
											
	pause_game = graphics_create_menu_item ( "_Pause Game",
											 "<gTuxBall>/Game/Pause Game",
											 NULL, menu, 
											 G_CALLBACK(&graphics_on_pause_game), NULL, FALSE, NULL	);
	
	gtk_image_menu_item_set_image (	GTK_IMAGE_MENU_ITEM( pause_game ),
									gtk_image_new_from_file ( "pixmaps/stock_timer_stopped.png" ) );
	
	graphics_menu_item_set_active( PAUSE_ITEM, 0 );
	
	graphics_create_menu_separator ( menu );
	
	menu_item = graphics_create_menu_item(	"_Scores...",
											"<gTuxBall-Main>/Game/Scores",
											NULL, menu,
											G_CALLBACK(&game_on_scores), NULL, FALSE, NULL );
											
	gtk_image_menu_item_set_image(	GTK_IMAGE_MENU_ITEM( menu_item ),
									gtk_image_new_from_file ( "pixmaps/stock_scores.png" ) );	
	
	end_game = graphics_create_menu_item(	"_End Game",
											"<gTuxBall>/Game/End Game",
											NULL, menu, G_CALLBACK(&graphics_on_end_game), NULL, FALSE, NULL );
	
	graphics_menu_item_set_active( END_ITEM, 0 );
	
	graphics_create_menu_separator( menu );
	
	menu_item = graphics_create_menu_item(	"_Quit Game",
											"<gTuxBall>/Game/Quit",
											GTK_STOCK_QUIT, menu,
											G_CALLBACK(&graphics_on_destroy), NULL, FALSE, NULL );
											


	/*=======================================================================*/
	/* Settings menu                                                         */
	/*=======================================================================*/
	settings_menu = gtk_menu_item_new_with_mnemonic( "_Settings" );
	
	
	menu = gtk_menu_new();
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(settings_menu), GTK_WIDGET(menu));
	
	gtk_menu_set_accel_group (GTK_MENU (menu), accel_group);
	gtk_accel_map_add_entry (	"<gTuxBall-Main>/Settings/Preferences",
								gdk_keyval_from_name ("p"),
								GDK_CONTROL_MASK);
	
	/*=======================================================================*/
	/* Level submenu                                                         */
	/*=======================================================================*/
	level_submenu = gtk_menu_item_new_with_mnemonic( "_Start Level" );
	
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), level_submenu);
	
	menu = gtk_menu_new();
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(level_submenu), GTK_WIDGET(menu));
	
	gtk_menu_set_accel_group (GTK_MENU (menu), accel_group);
	
	/* This variable contains a pointer to an array containing each level number,
	 * a pointer to the right item will be passed to the "Level x" menu item's
	 * callback.
	 */
	if( !( levels = ( gpointer )malloc( sizeof( int ) * NUMBER_OF_LEVELS ) ) )
		error( MEMORY_ERROR );
	
	/* For each level... */
	for( i = 0 ; i < NUMBER_OF_LEVELS ; i++ )
	{
		char str1[100], str2[2], str3[7];
		
		/* ...we fill in the right cell in the cell variable... */
		((int *)levels)[i] = i;
		
		/* ...we generate each needed string... */
		sprintf( str1, "<gTuxBall-Main>/Settings/Start Level/Level %d", i );
		sprintf( str2, "%d", i );
		sprintf( str3, "Level %d", i );
	
		/* ...add the accel map entry... */		
		gtk_accel_map_add_entry (	str1, gdk_keyval_from_name ( str2 ),
									GDK_CONTROL_MASK);
		
		
		/* ...create the menu item, associating the right cell's address
		 * (from level) to the callback... */
		menu_item = graphics_create_menu_item(	str3, str1, NULL, menu,
												G_CALLBACK(&graphics_on_level),
												levels + ( i * sizeof( int ) ),
												TRUE, group );
		
		/* ...and get its group to pass it to the following item. */
		group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (menu_item));
	}

	/*=======================================================================*/
	/* Help menu                                                             */
	/*=======================================================================*/
	help_menu = gtk_menu_item_new_with_mnemonic( "_Help" );
	
	menu = gtk_menu_new();
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(help_menu), GTK_WIDGET(menu));
	
	gtk_menu_set_accel_group (GTK_MENU (menu), accel_group);
	gtk_accel_map_add_entry (	"<gTuxBall-Main>/Help/About",
								gdk_keyval_from_name ("a"),
								GDK_CONTROL_MASK);
	
	menu_item = graphics_create_menu_item(	"_About",
											"<gTuxBall-Main>/Help/About",
											NULL, menu,
											(void *)(graphics_on_about), NULL, FALSE, NULL );
	

	gtk_image_menu_item_set_image(	GTK_IMAGE_MENU_ITEM( menu_item ),
									gtk_image_new_from_file ( "pixmaps/stock_menu_about.png" ) );	
	
	/* Finally, we append each menu to the menu bar. */
	gtk_menu_shell_append(GTK_MENU_SHELL (menu_bar), game_menu);
	gtk_menu_shell_append(GTK_MENU_SHELL (menu_bar), settings_menu);
	gtk_menu_shell_append(GTK_MENU_SHELL (menu_bar), help_menu);
}

/*
 * This functions allows to quickly create a menu item.
 */
GtkWidget *graphics_create_menu_item(	gchar * caption, gchar * accel,
										gchar * icon, GtkWidget * menu,
										GCallback callback, gpointer data,
										gboolean radio_menu_item, GSList * list )
{
	GtkWidget *menu_item, *pic;
	
	/* First, we create the menu item... */	
	if( radio_menu_item )
		menu_item = gtk_radio_menu_item_new_with_mnemonic( list, caption );
	else
		menu_item = gtk_image_menu_item_new_with_mnemonic( caption );
	
	/* ...then we set its accel path... */
	gtk_menu_item_set_accel_path( GTK_MENU_ITEM(menu_item), accel );
	/* ...set its icon... */
	if( icon && !radio_menu_item )
	{
		pic = gtk_image_new_from_stock( icon, GTK_ICON_SIZE_MENU );
		gtk_image_menu_item_set_image( GTK_IMAGE_MENU_ITEM(menu_item), pic );
	}
	/* ...insert it into the menu... */
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item);
	/* ...connect it with its handler... */
	if( callback ) g_signal_connect(	menu_item, "activate", 
										G_CALLBACK(callback), data );
	/* ...and then return it. */
	return menu_item;
}

/*
 * This functions allows to quickly create a menu separator.
 */
GtkWidget *graphics_create_menu_separator( GtkWidget * menu )
{
	/* We just create a NULL menu item... */
	GtkWidget *menu_item = gtk_menu_item_new();
	/* ...append it... */
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item);
	/* ...and return it. */
	return menu_item;
}


/*
 * This function is called when the window is destroyed.
 */
void graphics_on_destroy (GtkWidget * widget, gpointer data)
{
	int i;
	
	/* We leave the main event loop... */
    gtk_main_quit ();
	
	/* ...and then dereference/free the global objects. */
	free( levels );
	gdk_cursor_unref( arrow_cursor );
	gdk_cursor_unref( empty_cursor );
	g_object_unref( pad1_left_pixbuf );
	g_object_unref( pad2_left_pixbuf );
	g_object_unref( pad1_right_pixbuf );
	g_object_unref( pad2_right_pixbuf );
	g_object_unref( ball_pixbuf );
	g_object_unref( background_pixbuf );
	for( i = 1; i < NUMBER_OF_COLORS ; i++ )
		g_object_unref( blocks_pixbufs[i] );
}

/*
 * This function is called when the window has to be redrawn.
 */
gboolean graphics_on_expose_event ( GtkWidget *widget, GdkEventExpose *event,
						 			gpointer data)
{
	/* We do not need double-buffering here, as gtk handles it for us
	 * automatically. */
	GdkGC *gc = widget->style->fg_gc[GTK_WIDGET_STATE (widget)];
	GdkColor color;
	GtkImage *image;
	GdkPixbuf *pixbuf;
	
	int i,j;

	/* We first draw the background. */
	gdk_draw_pixbuf (	widget->window, gc, background_pixbuf, 0, 0, 0, 0,
						BOARD_WIDTH, BOARD_HEIGHT, GDK_RGB_DITHER_NONE, 0, 0);
	
	/* If there is a pad, draw it. */
	if( global_pad )
		(* ( element_get_draw_function( global_pad ) ) )( global_pad, widget );
	
	/* If there is a ball, draw it. */
	if( global_ball )
		(* ( element_get_draw_function( global_ball ) ) )( global_ball, widget );
	
	/* Now, we are going to loop through global_blocks in order to display every
	 * block with the graphics_draw_element function. */
	for( i = 0 ; i < NUMBER_OF_LINES ; i++ )
	{
		for( j = 0 ; j < NUMBER_OF_COLUMNS ; j++ )
		{
			if( global_blocks[i][j] != NULL )
			{
				(*(element_get_draw_function( global_blocks[i][j] )))( global_blocks[i][j], widget );
			}
		}
	}

	/* We display the right signal, depending on the game state. */
	if( global_state == WAITING_FOR_SIGNAL )
		graphics_draw_message( "Strike any key when ready", gc, widget, 1 );
	else if( global_state == PAUSE )
		graphics_draw_message( "Pause", gc, widget, 1 );
	else if( global_state == GAME_OVER )
		graphics_draw_message( "Game Over", gc, widget, 0 );
	else if( global_state == WON )
		graphics_draw_message( "You Win!", gc, widget, 0 );
	
	/* We do not need to propagate the event any further. */
	return TRUE;
}

/*
 * This function is called when the user presses a key.
 */
gboolean graphics_on_key_press_event( GtkWidget *widget,
              			              GdkEventKey *event,
                             		  gpointer user_data)
{
	/* We just call the right function in game.c */
	game_on_key_press_event( event->keyval );
	
	/* We do not need to propagate the event any further. */
	return FALSE;
}

/*
 * This function is called when the mouse moves.
 */
gboolean graphics_on_motion_notify_event ( GtkWidget *widget,
										   GdkEventMotion *event,
								  		   gpointer data )
{
	int x, y;
	
	/* We first get the mouse coordinates from the event. */
	x = event->x;
	y = event->y;
	
	/* If the event is an hint, we need to call gdk_window_get_pointer, in order
	   to get the next mouse events. We update x and y because these new
	   values are newer than those from the event structure. */
	if (event->is_hint) gdk_window_get_pointer(event->window, &x, &y, NULL);
	
	/* Then we call the right function in game.c */
	game_on_motion_notify_event( x, y );
	
	/* We do not need to propagate the event any further. */
	return TRUE;
}

/*
 * This function is called when the mouse enters the drawing area.
 */
gboolean graphics_on_enter_notify_event ( GtkWidget *widget,
										  GdkEventCrossing *event,
										  gpointer data )
{
	/* We hide the cursor. */
	gdk_window_set_cursor(widget->window, empty_cursor);
	
	/* We do not need to propagate the event any further. */
	return TRUE;
}

/*
 * This function is called when the mouse leaves the drawing area.
 */
gboolean graphics_on_leave_notify_event ( GtkWidget *widget,
										  GdkEventCrossing *event,
										  gpointer data )
{
	/* We show the cursor. */
	gdk_window_set_cursor( widget->window, arrow_cursor );
	
	/* We do not need to propagate the event any further. */
	return TRUE;
}

/*
 * This function is called when the user chooses "New Game" from the Game menu.
 */
void graphics_on_new_game ( GtkMenuItem *widget, gpointer user_data)
{
	game_on_new_game();
}

/*
 * This function is called when the user chooses "Pause Game" from the Game menu.
 */
void graphics_on_pause_game ( GtkMenuItem *widget, gpointer user_data)
{
	game_on_pause_game();
}

/*
 * This function is called when any of the "Scores" menu item is activated.
 */
void graphics_on_scores( GtkMenuItem *menuitem, gpointer data )
{
	graphics_show_high_scores();
}


/*
 * This function is called when the user chooses "End Game" from the Game menu.
 */
void graphics_on_end_game ( GtkMenuItem *widget, gpointer user_data )
{
	game_on_end_game();
}

/*
 * This function is called when any of the "Level" menu items is activated.
 */
void graphics_on_level( GtkMenuItem *menuitem, gpointer data )
{
	/* We call the associated callback from game, passing a pointer to an
	 * int containing the level number. */
	game_on_level( (*((int *)data)) );
}

/*
 * This function is called when the users chooses "About" in the help menu.
 */
void graphics_on_about( GtkMenuItem *widget, gpointer user_data )
{
	GtkWidget *dialog, *penguins, *label1, *label2, *label3;
	GtkStyle *style1, *style3;
	PangoFontDescription * font_desc = pango_font_description_new();
	
	/* We first create the dialog... */
	dialog = gtk_dialog_new_with_buttons ("About", GTK_WINDOW( window ),
                                         GTK_DIALOG_DESTROY_WITH_PARENT,
                                         GTK_STOCK_CLOSE,
                                         GTK_RESPONSE_NONE,
                                         NULL);

	/* ...and force its size. */
	gtk_widget_set_size_request (dialog,320,200);
	
	/* We load the logo into memory... */
	penguins = gtk_image_new_from_file ( LOGO_IMAGE_FILE );
	/* ...and insert it into the dialog. */
	gtk_container_add (	GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), 
						GTK_WIDGET( penguins ) );
	
	/* Then we create the first label. */
	label1 = gtk_label_new ( "gTuxBall v0.1" );
	label2 = gtk_label_new ( "Penguins are just so cool." );
	label3 = gtk_label_new ( "(C) Lozi Jean-Pierre, 2004 - http://gtuxball.lozi.org" );
	
	/* We store the label styles for label1 and label3, as we will change some
	 * of their style properties. */
	style1 = gtk_widget_get_style( label1 );
	style3 = gtk_style_copy( style1 );
	
	/* We update the font description for the first label, setting the text to
	 * Bold, 18 points. */
	pango_font_description_set_size( style1->font_desc, 18 * PANGO_SCALE );
	pango_font_description_set_weight( style1->font_desc, PANGO_WEIGHT_BOLD );
	gtk_widget_set_style( label1, style1 );
	
	/* We update the font description for the first label, setting the text to
	 * 8 points */
	pango_font_description_set_size( style3->font_desc, 8 * PANGO_SCALE );
	pango_font_description_set_weight( style1->font_desc, PANGO_WEIGHT_NORMAL );
	gtk_widget_set_style (label3, style3);
	
	/* Then we insert the 3 labels into the dialog. */
	gtk_container_add ( GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), label1 );
	gtk_container_add (	GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), label2 );
	gtk_container_add ( GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), label3 );

	/* When the user clicks on the OK button, the dialog is destroyed. */
   	g_signal_connect_swapped(	dialog, "response",
								G_CALLBACK (gtk_widget_destroy), dialog);

	/* Then we show the dialog and its contents... */				  
	gtk_widget_show_all ( dialog );
	
	/* ...and do not forget to free the PangoFontDescription, and to dereference
	 * the two styles. */	
	pango_font_description_free ( font_desc );
	g_object_unref( style1 );
	g_object_unref( style3 );
}


/*
 * This function wraps the real timeout function (in order to avoid to use GTK
 * types in game.c).
 */
gboolean graphics_timeout_function( gpointer data )
{
	/* We need to get the function result, because TRUE is defined as !0,
	 * which could be different from 1. */
	if ( (( Timeout_function )data )() == 1 ) return TRUE;
	else return FALSE;
}

/*
 * This function draws the given message.
 */
void graphics_draw_message( char * message, GdkGC * gc, GtkWidget * widget,
							int display_infos )
{
	PangoLayout *layout1, *layout2;
	PangoFontDescription *desc1, *desc2;
	GdkColor color;
	PangoRectangle *rect1, *rect2;
	GdkGCValues *values;
	int width;
	
	/* First, we store the GC values into a newly created structure. */
	if( !( values = ( GdkGCValues * )malloc( sizeof( GdkGCValues ) ) ) )
		error( MEMORY_ERROR );
	gdk_gc_get_values( gc, values );
	
	/* We first initialize the message font. */
	desc1 = pango_font_description_from_string ( "Luxi Sans 18" );
	/* This rectangle will contain the text extents, which allows us to
	 * center it. */
	rect1 = ( PangoRectangle* )malloc( sizeof( PangoRectangle ) );
	/* We set the right font to the layout. */
	layout1 = gtk_widget_create_pango_layout ( widget, message );
	pango_layout_set_font_description( layout1, desc1 );
	pango_layout_get_pixel_extents( layout1, rect1, NULL );

	if( display_infos )
	{
		char str[32];
		/* We create a string containing the game informations. */
		sprintf( str, "Level %d - %d Live%s Left",
					  global_current_level, 
					  global_lives, ( ( global_lives > 1 ) ? "s" : "" ) );
		/* We first initialize the message font. */
		desc2 = pango_font_description_from_string ( "Luxi Sans 14" );
		/* This rectangle will contain the text extents, which allows us to
		 * center it. */
		rect2 = ( PangoRectangle* )malloc( sizeof( PangoRectangle ) );
		/* We set the right font to the layout. */
		layout2 = gtk_widget_create_pango_layout ( widget, str );
		pango_layout_set_font_description( layout2, desc2 );
		pango_layout_get_pixel_extents( layout2, rect2, NULL );
	}
	
	/* We set the color to black, in order to draw the text shadow. */
	gdk_color_parse( "#000000", &color );
    gdk_gc_set_rgb_fg_color( gc, &color);
	
	/* If we need to display the game informations... */
	if( display_infos )
	{
		/* We first draw the game informations, then the given message. */
		gdk_draw_layout (	widget->window, gc,
							BOARD_WIDTH/2-( ( rect1->width ) / 2 ) + 1, 
							BOTTOM_MESSAGE_Y + 1, layout1 );
		gdk_draw_layout (	widget->window, gc,
							BOARD_WIDTH/2-( ( rect2->width ) / 2 ) + 1, 
							TOP_MESSAGE_Y + 1, layout2 );
	}
	/* Otherwise... */
	else
	{
		/* We just draw the message. */
		gdk_draw_layout (	widget->window, gc,
							BOARD_WIDTH/2-( ( rect1->width ) / 2 ) + 1, 
							TOP_MESSAGE_Y + 1, layout1 );
	}
	
	/* Then we set the color to white, to draw the text itself. */
	gdk_color_parse ("#FFFFFF", &color);
    gdk_gc_set_rgb_fg_color (gc, &color);
	
	/* If we need to display the game informations... */
	if( display_infos )
	{
		/* We first draw the game informations, then the given message. */
		gdk_draw_layout (	widget->window, gc,
							BOARD_WIDTH/2-( ( rect1->width ) / 2 ) + 1, 
							BOTTOM_MESSAGE_Y, layout1 );
		gdk_draw_layout (	widget->window, gc,
							BOARD_WIDTH/2-( ( rect2->width ) / 2 ) + 1, 
							TOP_MESSAGE_Y, layout2 );
	}
	/* Otherwise... */
	else
	{
		/* We just draw the message. */
		gdk_draw_layout (	widget->window, gc,
							BOARD_WIDTH/2-( ( rect1->width ) / 2 ) + 1, 
							TOP_MESSAGE_Y, layout1 );
	}
	
	/* Then we restore the gc's values. */
	gdk_gc_set_values( gc, values, GDK_GC_FOREGROUND);
	
	/* And free the allocated data. */
	pango_font_description_free( desc1 );
	g_object_unref( layout1 );
	free( rect1 );
	if( display_infos )
	{
		pango_font_description_free( desc2 );
		g_object_unref( layout2 );
		free( rect2 );
	}
	free( values );
}

/*
 * This function invalidates the whole graphic area.
 */
void graphics_invalidate_all( void )
{
	GdkRectangle *rect;
	
	/* We create a new rectangle containing the whole area... */
	rect = (GdkRectangle *)malloc( sizeof(GdkRectangle) );
	rect->x = 0;
	rect->y = 0;
	rect->width = BOARD_WIDTH;
	rect->height = BOARD_HEIGHT;
	
	/* ...call the right function from the GTK+ API... */
	gdk_window_invalidate_rect( GTK_WIDGET( drawing_area )->window, rect, TRUE);
	
	/* ...and then free the rectangle. */
	free( rect );
}

/*
 * This function invalidates a part of the graphic area.
 */
void graphics_invalidate_rect( 	int top_left_x, int top_left_y,
								int bottom_right_x, int bottom_right_y )
{
	GdkRectangle *rect;
	
	/* We create a new rectangle containing the whole area... */
	rect = (GdkRectangle *)malloc( sizeof(GdkRectangle) );
	rect->x = top_left_x;
	rect->y = top_left_y;
	rect->width = bottom_right_x - top_left_x;
	rect->height = bottom_right_y - top_left_y;
	
	/* ...call the right function from the GTK+ API... */
	gdk_window_invalidate_rect( GTK_WIDGET( drawing_area )->window, rect, TRUE);
	
	/* ...and then free the rectangle. */
	free( rect );
}

/*
 * This function allows to set the active state of a menu item.
 */
void graphics_menu_item_set_active( int item, int active )
{
	/* We just enable/disable the right item. */
	switch( item )
	{
		case PAUSE_ITEM:
			gtk_widget_set_sensitive (GTK_WIDGET(pause_game),active);
			break;
		
		case END_ITEM:
			gtk_widget_set_sensitive (GTK_WIDGET(end_game),active);
			break;
	}
}

/*
 * This function, when called, displays a prompt asking the user's name.
 */
void graphics_ask_name ( char * name )
{
	GtkWidget *dialog, *label, *entry;
	
	/* We create the modal dialog. */
	dialog = gtk_dialog_new_with_buttons( "High score!", GTK_WINDOW( window ),
                                          GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                                          GTK_STOCK_OK, GTK_RESPONSE_OK,
										  GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL );
	
	/* We create the label. */
	label = gtk_label_new ( "Please enter your name :" );
	
	/* We create the 'entry' : an editable text area where the user can
	 * type his name. */
	entry = GTK_WIDGET( gtk_entry_new_with_max_length( MAX_NAME_LENGTH ) );
	
	/* Then, we insert the two widgets into the dialog... */
	gtk_container_add (	GTK_CONTAINER ( GTK_DIALOG( dialog )->vbox ), label );
	gtk_container_add (	GTK_CONTAINER ( GTK_DIALOG( dialog )->vbox ), entry );
	
	/* ...show them... */
	gtk_widget_show( label );
	gtk_widget_show( entry );
	
	/* ...then run the dialog. */
	switch (gtk_dialog_run(GTK_DIALOG(dialog)))
    {
		/* If the user chose the OK button, we save his name. */
        case GTK_RESPONSE_OK:
            strcpy(name, gtk_entry_get_text(GTK_ENTRY(entry)));
            break;
		/* Otherwise, we return the default name. */
        default:
			strcpy(name, DEFAULT_USER_NAME);
            break;
    }
	
	/* Finally, we destroy the dialog. */
	gtk_widget_destroy( dialog ); 
}


/* 
 * This function shows the "High Scores" dialog box.
 */
void graphics_show_high_scores( void )
{
	GtkWidget *dialog, *penguins, *label1, *label2, *label3;
	GtkWidget *hbox, *vbox, *label;
	int i = 0, column;
	char str[64], format[26], date_time[100];
	struct tm *time, *currtime;
	time_t seconds, ltime;
	Score * scores = score_read_scores();
	char * column_names[] = { "Name", "Blocks", "Start Level",
							  "Level", "Game Time", "Date" };
	
	/* We first create the dialog... */
	dialog = gtk_dialog_new_with_buttons ( "Top Ten", GTK_WINDOW( window ),
                                           GTK_DIALOG_DESTROY_WITH_PARENT,
                                           GTK_STOCK_CLOSE,
                                           GTK_RESPONSE_CLOSE,
                                           NULL );
	
	
	/* First we create a label containing the app's name... */
	label = gtk_label_new( "gTuxBall" );
	/* ...and insert it into the dialog. */
	gtk_container_add (	GTK_CONTAINER ( GTK_DIALOG( dialog )->vbox ), label );
	
	/* Then we create an horizontal box which will contain the columns... */
	hbox = gtk_hbox_new( FALSE, 5 );
	/* ...and insert it into the dialog. */
	gtk_container_add (	GTK_CONTAINER ( GTK_DIALOG( dialog )->vbox ), hbox );
	

	for( column = 0 ; column < 6 ; column++ )
	{
		/* We first create the vbox... */
		vbox = gtk_vbox_new( TRUE, 5 );
		/* ...and add it to the horizontal box. */
		gtk_container_add (	GTK_CONTAINER ( hbox ), vbox );
		/* We then create the label containing the column name... */
		label = gtk_label_new( column_names[column] );
		/* ...and add it to the horizontal box. */
		gtk_container_add (	GTK_CONTAINER ( vbox ), label );
		/* Finally, we insert each field. */
		for( i = 0; i < 10; i++ )
		{
			/* The way each field is shown depends on the column. */
			switch( column )
			{
				case SCORE_NAME :
					/* We add to the container... */
					gtk_container_add (	GTK_CONTAINER ( vbox ),
						/* ...a label... */	
						gtk_label_new( ( char * )score_get_attribute( 
							/* ...containing the score name. */
							scores[i], SCORE_NAME )));
				break;
				
				case SCORE_LEVEL :
				case SCORE_START_LEVEL :
				case SCORE_NUMBER_OF_BLOCKS :
					/* We need to print the number into a string in order to display it... */
					sprintf( str, "%d", ( int )score_get_attribute( scores[i], column ) );
					/* ...then we create the label. */
					gtk_container_add (	GTK_CONTAINER ( vbox ), gtk_label_new( str ) );
					break;
				
				case SCORE_DIFF_TIME :
					/* We fist get the diff time, and format it. */
					strcpy( format, "%H:%M:%S" );
					seconds = ( time_t )score_get_attribute( scores[i], SCORE_DIFF_TIME );
					time = gmtime( &seconds );
					strftime ( date_time, 26, format, time);
					/* Then we create a new label... */
					label = gtk_label_new( date_time );
					/* ...and insert it into the column. */
					gtk_container_add (	GTK_CONTAINER ( vbox ), label );
					break;
				
				case SCORE_START_TIME :
					/* We fist get the diff time, and format it. */
					strcpy( format, "%a %d %m %H:%M:%S %Y" );
					ltime = ( time_t )score_get_attribute( scores[i], SCORE_START_TIME );
					currtime = localtime( &ltime );
					strftime ( date_time, 26, format, currtime);
					/* Then we create a new label... */
					label = gtk_label_new( date_time ); 
					/* ...and insert it into the column. */
					gtk_container_add (	GTK_CONTAINER ( vbox ), label );
					break;
			}
		}
	}
	
	/* When the user clicks on the OK button, the dialog is destroyed. */
   	g_signal_connect_swapped(	dialog, "response",
								G_CALLBACK (gtk_widget_destroy), dialog);
	
	/* Then we show the widget... */
	gtk_widget_show_all( dialog );
	
	/* ...and free the allocated memory. */
	for( i = 9; i >= 0 ; i-- ) free( scores[i] );
	free( scores );
}

/*
 * This function draws the given element.
 */
void graphics_draw_block( int color, int x, int y, int width, int height, void * data )
{
	/* We just draw the pixbuf. */
	gdk_draw_pixbuf( GTK_WIDGET( data )->window,
					 GTK_WIDGET( data )->style->fg_gc[GTK_WIDGET_STATE (data)],
					 blocks_pixbufs[color], 0, 0, x, y, width, height,
					 GDK_RGB_DITHER_NONE, 0, 0);
}

/*
 * This function draws the ball at the given coordinates (with the given width and
 * height).
 */
void graphics_draw_ball( int x, int y, int width, int height, void * data )
{
	/* We just draw the pixbuf. */
	gdk_draw_pixbuf( GTK_WIDGET( data )->window,
					 GTK_WIDGET( data )->style->fg_gc[GTK_WIDGET_STATE (data)],
					 ball_pixbuf, 0, 0, x, y, width, height, GDK_RGB_DITHER_NONE, 0, 0 );
}

/*
 * This function draws the pad at the given coordinates (with the given width and
 * height).
 */
void graphics_draw_pad( int x, int y, int width, int height, void * data )
{
	GdkPixbuf *pixbuf;
	
	/* We choose the right picture, depending on the direction of the pad,
	 * and the feet state. */
	if( global_feet && global_direction == LEFT )
		pixbuf = pad1_left_pixbuf;
	else if( global_feet && global_direction == RIGHT )
		pixbuf = pad1_right_pixbuf;
	else if( !global_feet && global_direction == LEFT ) pixbuf =
		pad2_left_pixbuf;
	else
		pixbuf = pad2_right_pixbuf;
	
	/* Then we draw the pixbuf. */
	gdk_draw_pixbuf ( GTK_WIDGET( data )->window,
					  GTK_WIDGET( data )->style->fg_gc[GTK_WIDGET_STATE (data)],
					  pixbuf, 0, 0, x, y, width, height, GDK_RGB_DITHER_NONE, 0, 0);
}

/*
 * Calling this function will make the on_timer function called each timeout seconds.
 */
void graphics_add_timeout( int timeout, Timeout_function on_timer )
{
	g_timeout_add( timeout, &graphics_timeout_function, on_timer );
}
