/***************************************************************************
 *            score.c
 *
 *  Fri Dec 24 18:37:10 2004
 *  Copyright  2004  Jean-Pierre Lozi
 *  jean-pierre@lozi.org - http://gtuxball.lozi.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>

#include "error.h"
#include "globals.h"
#include "score.h"

/* The score structure... */
struct score {
	/* ...contains the player name... */
	char name[64];
	/* ...the number of blocks, level, start level... */
	int number_of_blocks, level, start_level;
	/* ...the start time... */
	long start_time;
	/* ...and the game time. */
	long diff_time;
};

/*
 * This function creates and returns a score.
 */
Score score_create( char * name, int number_of_blocks, int start_level,
					int level, time_t diff_time, time_t start_time )
{
	Score output;
	
	/* We allocate a new score structure... */
	if( !(output = ( Score )malloc( sizeof( struct score ) ) ) )
		error( MEMORY_ERROR );
	
	/* ...we initialize each field... */
	strcpy( output->name, name );
	output->number_of_blocks = number_of_blocks;
	output->start_level = start_level;
	output->level = level;
	output->diff_time = diff_time;
	output->start_time = start_time;
	
	/* ...then we return a pointer to the newly allocated memory */
	return output;
}

/*
 * This function destroys the given score.
 */
void score_destroy( Score *obj )
{
	/* First, we free the element... */
	free( *obj );
	/* ...and then update the pointer. */
	*obj = NULL;
}

/*
 * This function returns the specified attribute.
 */
long score_get_attribute( Score scr, int attribute )
{
	/* We return each attribute as a long, we will need to typecast
	 * them after the function is called. */
	switch( attribute )
	{
		case SCORE_NAME:
			return ( long )( scr->name );
		case SCORE_NUMBER_OF_BLOCKS:
			return ( long )( scr->number_of_blocks );
		case SCORE_START_LEVEL:
			return ( long )( scr->start_level );
		case SCORE_LEVEL:
			return ( long )( scr->level );
		case SCORE_DIFF_TIME:
			return ( long )( scr->diff_time );
		case SCORE_START_TIME:
			return ( long )( scr->start_time );
	}
}

/*
 * This function reads the scores from the high scores file, and stores them
 * into an array.
 */
Score * score_read_scores ( void )
{
	FILE *file;
	char file_contents[MAX_SCORE_STORED_SIZE * NUMBER_OF_SCORES];
	char * game, *ptr;
	Score * scores;
	int i = 0;
	int j = 0;
	long z;
	char name[MAX_NAME_LENGTH];
	
	/* We first allocate the array. */
	if( !( scores = ( Score * )malloc( NUMBER_OF_SCORES * sizeof( Score ) ) ) )
		error( MEMORY_ERROR );
	
	/* Then, we open the high scores file. */
	file = fopen( HIGH_SCORES_FILE, "r");
	
	/* If it does not exist...*/
	if( file == NULL )
	{
		int i;
		/* ...then we reopen it in the write mode... */
		file = fopen( HIGH_SCORES_FILE, "w");
		
		/* If we can't open the file, we call the error function. */
		if( file == NULL ) error( HIGH_SCORES_OPEN_FILE_ERROR );
		
		/* ...fill it in with the default values... */
		for( i = 0; i < NUMBER_OF_SCORES; i++ ) 
			fprintf( file, "%s:0:0:0:0:0:", DEFAULT_USER_NAME );
		fprintf(file, "\n");
		
		/* ...and reopen it in the read mode. */
		file = freopen( HIGH_SCORES_FILE, "r",file);
	}
	
	/* Then we get the file's contents. */
	fgets( file_contents, MAX_SCORE_STORED_SIZE * NUMBER_OF_SCORES, file );
	
	ptr = file_contents;
	i = 0;
	
	/* For each value in the file... */
	while( ( game = (char *)strtok( ptr, ":" ) ) != NULL )
	{
		/* ...we store it in the right field of the right cell. */
		switch( j % 6 )
		{
			case 0:
				/* We allocate each score when we need to fill it in. */
				if( !(scores[i] = ( Score )malloc( sizeof( struct score ) ) ) )
					error( MEMORY_ERROR );
				/* We store the player's name... */
				strcpy( scores[i]->name, game );
				break;
			case 1:
				/* We store the number of blocks... */
				sscanf(game,"%d", &( scores[i]->number_of_blocks ) );
				break;
			case 2:
				/* We store the start level... */
				sscanf(game,"%d",&( scores[i]->start_level ) );
				break;
			case 3:
				/* We store the level... */
				sscanf(game,"%d", &( scores[i]->level ) );
				break;
			case 4:
				/* We store the diff_time... */
				sscanf(game,"%d", &( scores[i]->diff_time ) );
				break;
			case 5:
				/* We store the start time... */
				sscanf(game,"%d", &( scores[i]->start_time ) );
				i++;
				break;
		}
		j++;
		/* If the file contains more values than needed, we just ignore them. */
		if( i == NUMBER_OF_SCORES ) break;
		ptr = NULL;
	}
	
	/* If we got too few values, we call the error function. */
	if( j < NUMBER_OF_SCORES * NUMBER_OF_FIELDS )
		error( HIGH_SCORES_FILE_ERROR );
	
	/* Then we close the file... */
	fclose( file );
	
	/* ...and return the array. */
	return scores;
}
